import datetime
import json
import re
from typing import Any

from .log_processor import LogProcessor


class JSONComparator:
    """
    Provides comparison of two JSON-files.
    """

    def __init__(
        self,
        left_file_path: str,
        right_file_path: str,
        key: str | None = None,
        ignore: str | list[str] | None = None,
        ignore_types: bool = False,
    ):
        """
        :param left_file_path: path to file in .JSON format;
        :param right_file_path: path to file in .JSON format;
        :param key: key for JSON-object's property's name to be a key in comparison
        of similar objects nested in an array. For example, you have:
        {"cats": [{"id": 4, "name": "Nyan"}, {"id": 2, "name": "Marx"}]}. If you want to set
        cat's "id" as a key, use "DATA.cats.<array>.id". DATA points out to the root
        of your JSON and <array> indicates object with key is nested in an array;
        :param ignore: a str or a list of objects' properties whose mismatch should be
        ignored in the process of comparing. Checkout key's paragraph to see examples
        of correct setting the param's items;
        :param ignore_types: set True if you want JSONComparator to ignore type differences
        between similar values like "1.4" vs 1.4 or "[\"New Age №1\"]" vs ["New Age №1"].
        """
        self.diff_log: LogProcessor = LogProcessor()
        self.key: list[str] = list()
        if isinstance(key, list):
            self.key = key
        elif isinstance(key, str):
            self.key = [key]
        self.__match_keys: list[str] = list()
        self.ignore: list[str] = list()
        if isinstance(ignore, list):
            self.ignore = ignore
        elif isinstance(ignore, str):
            self.ignore = [ignore]
        self.ignore_types: bool = ignore_types
        with open(left_file_path, "r") as file:
            self.left_data = json.load(file)
            self.left_f_name: str = file.name
        with open(right_file_path, "r") as file:
            self.right_data = json.load(file)
            self.right_f_name: str = file.name

    def compare_with_right(self) -> None:
        """
        Looks for how the right JSON-file differs from the left one.
        :return: None;
        """
        self.__clear_temp_data()
        self.__edit_key_path_root("DATA", self.right_f_name)
        self.__compare()
        self.__edit_key_path_root(self.right_f_name, "DATA")
        self.diff_log._setup_summary()

    def compare_with_left(self) -> None:
        """
        Looks for how the left JSON-file differs from the right one.
        :return: None;
        """
        self.__clear_temp_data()
        self.__edit_key_path_root("DATA", self.left_f_name)
        self.__compare(with_right=False)
        self.__edit_key_path_root(self.left_f_name, "DATA")
        self.diff_log._setup_summary()

    def full_compare(self) -> None:
        """
        Looks for differences between files from both perspectives.
        :return: None;
        """
        self.__clear_temp_data()
        self.__edit_key_path_root("DATA", self.right_f_name)
        self.__compare()
        self.__edit_key_path_root(self.right_f_name, self.left_f_name)
        self.__compare(with_right=False)
        self.__edit_key_path_root(self.left_f_name, "DATA")
        self.diff_log._setup_summary()

    def save_diff_logs(self, path: str = "") -> None:
        """
        Saves recorded logs as a text file.
        :param path: a path to directory where the log with differences will be saved.
        By default, current work directory will be used;
        :return: None;
        """
        log_file_name = f"json_compare_diff_{datetime.date.today()}"
        with open(f"{path}{log_file_name}", "w") as f:
            f.write("\n".join(self.diff_log.log))

    def __edit_key_path_root(self, from_: str, to: str) -> None:
        self.key = [key.replace(from_, to) for key in self.key]

    def __compare(self, with_right: bool = True) -> None:
        data1, data2, root_log = self.left_data, self.right_data, self.right_f_name
        if not with_right:
            data1, data2, root_log = data2, data1, self.left_f_name

        if isinstance(self.left_data, list) and isinstance(self.right_data, list):
            self.__compare_lists(root_log, data1, data2)
        elif isinstance(self.left_data, dict) and isinstance(self.right_data, dict):
            self.__compare_dicts(root_log, data1, data2)
        else:
            self.diff_log._incorrect_type(self.left_data, self.right_data)

    def __compare_dicts(
        self, item_path: str, exp_data: dict[str, Any], act_data: dict[str, Any]
    ) -> None:
        for key, val in exp_data.items():
            self.diff_log._setup_path(item_path, key)
            if key not in act_data:
                self.diff_log._missing_property()
                continue
            if type(val) is dict:
                self.__try_find_dict_and_compare(val, act_data[key])
                continue
            if type(val) is list:
                self.__try_find_list_and_compare(val, act_data[key])
                continue
            if val != act_data[key] and not self.__key_to_ignore():
                self.__compare_values(val, act_data[key])

    def __compare_lists(
        self, item_path: str, exp_data: list[Any], act_data: list[Any]
    ) -> None:
        self.diff_log._setup_path(item_path, "<array>")
        self.__setup_match_keys()
        if self.__match_keys:
            self.__compare_lists_by_keys(self.diff_log.curr_path, exp_data, act_data)
        else:
            self.__compare_lists_by_order(self.diff_log.curr_path, exp_data, act_data)

    def __compare_lists_by_order(
        self, item_path: str, exp_data: list[Any], act_data: list[Any]
    ) -> None:
        exp_data_len, act_data_len = len(exp_data), len(act_data)
        self.__check_array_lengths(item_path, exp_data_len, act_data_len)

        for idx, val in enumerate(exp_data):
            self.diff_log._setup_path(item_path, f"[{idx}]")
            if idx >= act_data_len:
                break
            if type(val) is dict:
                self.__try_find_dict_and_compare(val, act_data[idx])
                continue
            if type(val) is list:
                self.__try_find_list_and_compare(val, act_data[idx])
                continue
            if val != act_data[idx]:
                self.__compare_values(val, act_data[idx])

    def __compare_lists_by_keys(
        self, item_path: str, exp_data: list[Any], act_data: list[Any]
    ) -> None:
        exp_data_len, act_data_len = len(exp_data), len(act_data)
        self.__check_array_lengths(item_path, exp_data_len, act_data_len)

        for idx, val in enumerate(exp_data):
            self.diff_log._setup_path(item_path, f"[{idx}]")
            if type(val) is dict:
                target_keys_with_values = self.__get_key_props_with_values(val)
                act_data_similar_idx = self.__get_idx_of_similar_row(
                    target_keys_with_values, act_data
                )
                if act_data_similar_idx is None:
                    self.diff_log.missing_array_item(target_keys_with_values)
                    continue
                self.__compare_dicts(self.diff_log.curr_path, val, act_data[act_data_similar_idx])
                continue
            if type(val) is list:
                self.__try_find_list_and_compare(val, act_data[idx])
                continue
            if val != act_data[idx]:
                self.__compare_values(val, act_data[idx])

    def __try_find_list_and_compare(
        self,
        expected_val: list, 
        actual_val: int | float | str | bool | list[Any] | dict[str, Any] | None,
    ) -> None:
        if isinstance(actual_val, list):
            self.__compare_lists(self.diff_log.curr_path, expected_val, actual_val)
        elif isinstance(actual_val, str) and self.ignore_types:
            self.__try_json_decode_and_compare(expected_val,actual_val)

    def __try_find_dict_and_compare(
        self,
        expected_val: dict, 
        actual_val: int | float | str | bool | list[Any] | dict[str, Any] | None,
    ) -> None:
        if isinstance(actual_val, dict):
            self.__compare_dicts(self.diff_log.curr_path, expected_val, actual_val)
        elif isinstance(actual_val, str) and self.ignore_types:
            self.__try_json_decode_and_compare(expected_val, actual_val)

    def __check_array_lengths(self, exp_len: int, act_len: int) -> None:
        if exp_len > act_len:
            self.diff_log._lack_of_array_items(exp_len, act_len)
        elif exp_len < act_len:
            self.diff_log._exceeding_array_items(exp_len, act_len)

    def __get_key_props_with_values(
        self, object_: dict[str, Any]
    ) -> dict[str, int | float | str | bool | dict[str, Any] | list[Any] | None]:
        props_with_values = dict()
        for key in self.__match_keys:
            props_with_values[key] = object_[key]
        return props_with_values

    def __key_to_ignore(self) -> bool:
        curr_path_without_idx = (
            re.sub("\[[^()]*\]", "", self.diff_log.curr_path)
            .replace(self.right_f_name, "DATA")
            .replace(self.left_f_name, "DATA")
            .replace("////", "//")
        )
        return True if curr_path_without_idx in self.ignore else False

    @staticmethod
    def __get_idx_of_similar_row(
        keys_with_values: dict[str, int | float | str | bool | dict[str, Any] | list[Any] | None],
        data: list[dict[str, Any]],
    ) -> int | None:
        for idx, obj in enumerate(data):
            matched_props = [True for k, v in keys_with_values.items() if obj[k] == v]
            if len(matched_props) == len(keys_with_values):
                return idx
        return None

    def __compare_values(
        self,
        expected_val: int | float | str | bool | None,
        actual_val: int | float | str | bool | None,
    ) -> None:
        if expected_val == actual_val:
            return
        if not self.ignore_types:
            return self.diff_log._unequal_values(expected_val, actual_val)
        try:
            if isinstance(expected_val, int) and expected_val == int(actual_val):  # type: ignore
                return
            if isinstance(expected_val, float) and expected_val == float(actual_val):  # type: ignore
                return
            if isinstance(expected_val, str) and expected_val == str(actual_val):
                return
            if isinstance(expected_val, bool) and expected_val == bool(actual_val):
                return
        except ValueError:
            pass
        self.diff_log._unequal_values(expected_val, actual_val)
    
    def __try_json_decode_and_compare(
        self,
        expceted_val: dict[str, Any] | list[Any],
        actual_val: str,
    ) -> None:
        if isinstance(expceted_val, dict):
            try:
                parsed_val = json.loads(actual_val)
                if isinstance(parsed_val, dict):
                    self.__compare_dicts(self.diff_log.curr_path, expceted_val, parsed_val)
            except TypeError:
                self.diff_log._incorrect_type(expceted_val, actual_val)
        else:
            try:
                parsed_val = json.loads(actual_val)
                if isinstance(parsed_val, list):
                    self.__compare_lists(self.diff_log.curr_path, expceted_val, parsed_val)
            except TypeError:
                self.diff_log._incorrect_type(expceted_val, actual_val)

    def __setup_match_keys(self) -> None:
        if not self.key:
            return None
        match_keys = []
        for key in self.key:
            key_without_path = key.removeprefix(self.diff_log.curr_path + "//")
            if "//" not in key_without_path:
                match_keys.append(key_without_path)
        self.__match_keys = match_keys

    def __clear_temp_data(self) -> None:
        self.diff_log.log.clear()
        self.__match_keys.clear()
        self.diff_log.curr_path = str()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['json_compare']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'json-files-compare',
    'version': '1.0.1',
    'description': 'Simple but powerfull JSON-files comparator',
    'long_description': '# json-compare\nJson-compare is a simple package that allows you to easily and fastly compare two .json files. Support key and multi-key comparison.\nYou can also ignore certain fields\' values or perform comparison insensitive to data types.\n\n[![made-with-python](https://img.shields.io/badge/python-3.10%2B-brightgreen)](https://www.python.org/)\n\nUsage\n---\nCompare files just as they are:\n\n```python\nfrom json_compare import JSONComparator\n\ncomparator = JSONComparator(\n    left_file_path="expected.json",\n    right_file_path="actual.json",\n)\n\n# compare "actual.json" from the perspective of "expected.json"\'s structure\ncomparator.compare_with_right()  # / compare_with_left() / full_compare()\n\n# save diff logs to ".comparison_logs" folder\ncomparator.save_diff_logs(path="comparison_logs")\n\n# or print them into stdout\ndiffs = comparator.diff_log.log\nprint("\\n".join(diffs))\n\n# or print only summary. Here\'s an example:\nprint(comparator.diff_log.get_summary())\n\n# OUTPUT example:\n# ---------------------\n# TOTAL: 4 differences\n# -missing_obj_property: 3\n# -unequal_value: 4\n```\nSet key property to perform more accurate comparisons of objects in arrays:\n\n```python\n# expected.json: {"cats": [{"id": 4, "name": "Nyan"}, {"id": 2, "name": "Marx"}, {"id": 8, "name": "Flake"}]}\n# actual.json: {"cats": [{"id": 2, "name": "Marx"}, {"id": 4, "name": "Naan"}]}\n\ncomparator = JSONComparator(\n    left_file_path="expected.json",\n    right_file_path="actual.json",\n    key="DATA//cats//<array>//id",  # <----- just pass a "path" to needed property with following keywords: \n)                                            # DATA - points to the root of file \n                                             # <array> - indicates array with key property\'s object\n```\nIn this case, saved diff log would look like that:\n```text\nactual.json//cats//<array>\nlack of items in array: expected 3 items, got only 2\nactual.json//cats//<array>//[0]//name\nunequal values: expected "Nyan", got "Naan" instead\nactual.json//cats//<array>//[2]\nmissing array item: expected <object> with "id"=8\n```\nYou can go further and add non-important fields to `ignore` parameter:\n```python\n# expected.json: [{"id": 4, "name": "Nyan", "age": 2}, {"id": 2, "name": "Marx", "age": 7}, {"id": 8, "name": "Flake", "age": 4}]\n# actual.json: [{"id": 2, "name": "Marx", "age": 7}, {"id": 4, "name": "Naan", "age": "two"}, {"id": 9, "name": "Lol", "age": 1}]\n\ncomparator = JSONComparator(\n    left_file_path="expected.json",\n    right_file_path="actual.json",\n    key="DATA//<array>//id",\n    ignore="DATA//<array>//age"  # <-------\n)  \n```\nAnd here the result:\n```text\nactual.json//<array>//[0]//name\nunequal values: expected "Nyan", got "Naan" instead\nactual.json//<array>//[2]\nmissing array item: expected <object> with "id"=8\n```\nIf you want to compare ignoring type-differences between similar values\n like `"1.4"` vs `1.4` or `"[\\"New Age №1\\"]"` vs `["New Age №1"]` - just add `ignore_types=True` \n param to JSONComparator:\n```python\ncomparator = JSONComparator(\n    left_file_path="expected.json",\n    right_file_path="actual.json",\n    key="DATA//<array>//id",\n    ignore_types=True,  # <-------\n)  \n```\n',
    'author': 'Alexey Novikov',
    'author_email': 'lexutka@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from collections import Counter
from typing import Any


class LogProcessor:
    def __init__(self) -> None:
        self.log: list[str] = list()
        self.curr_path: str = str()
        self.diffs_counter: Counter[Any] = Counter(
            {
                "missing_obj_property": 0,
                "incorrect_type": 0,
                "arr_with_lack_of_items": 0,
                "exceeding_array_items": 0,
                "unequal_value": 0,
                "missing_array_item": 0,
            }

        )
        
    def get_summary(self) -> str | None:
        """
        :return: if comparison was performed, with summary differences in format:
        ---------------------
        TOTAL: 4 differences
        -missing_obj_property: 3
        -unequal_value: 4
        """
        if hasattr(self, "summary"):
            return self.summary
        return None

    def _setup_path(self, prev_path: str, key: str) -> None:
        if prev_path:
            self.curr_path = prev_path + f"//{key}"
        else:
            self.curr_path = f"//{key}"

    def _missing_property(self) -> None:
        msg = self.curr_path + "\nproperty is missing"
        self.log.append(msg)
        self.diffs_counter["missing_obj_property"] += 1

    def _incorrect_type(
            self,
            exp_obj: int | float | str | bool | dict[str, Any] | list[Any] | None,
            act_obj: int | float | str | bool | dict[str, Any] | list[Any] | None,
    ) -> None:
        exp_obj_type = self.__convert_to_json_type(exp_obj)
        act_obj_type = self.__convert_to_json_type(act_obj)
        msg = (
            self.curr_path + f"\nincorrect type: expected {exp_obj} ({exp_obj_type}), "
            f"got {act_obj} ({act_obj_type}) instead"
        )
        self.log.append(msg)
        self.diffs_counter["incorrect_type"] += 1

    def _lack_of_array_items(self, exp_len: int, act_len: int) -> None:
        msg = (
            self.curr_path
            + f"\nlack of items in array: expected {exp_len} items, got only {act_len}"
        )
        self.log.append(msg)
        self.diffs_counter["arr_with_lack_of_items"] += 1

    def _exceeding_array_items(self, exp_len: int, act_len: int) -> None:
        msg = (
            self.curr_path
            + f"\ntoo much items in array: expected {exp_len} items, got {act_len}"
        )
        self.log.append(msg)
        self.diffs_counter["exceeding_array_items"] += 1

    def _unequal_values(
        self,
        exp_value: int | float | str | bool | None,
        act_value: int | float | str | bool | None,
    ) -> None:
        if type(exp_value) != type(act_value):
            return self._incorrect_type(exp_value, act_value)

        if isinstance(exp_value, str):
            exp_value = f'"{exp_value}"'
            act_value = f'"{act_value}"'
        msg = (
            self.curr_path
            + f"\nunequal values: expected {exp_value}, got {act_value} instead"
        )
        self.log.append(msg)
        self.diffs_counter["unequal_value"] += 1

    def missing_array_item(
        self,
        key_props_with_values: (
                dict[str, int | float | str | bool | dict[str, Any] | list[Any] | None]
        ),
    ) -> None:
        expected_props = list()
        for key, val in key_props_with_values.items():
            if isinstance(val, str):
                val = f'"{val}"'
            expected_props.append(f'{key}: {val}')
        expected_props_str = ', '.join(expected_props)
        msg = self.curr_path + f"\nmissing array item: expected <object> with {expected_props_str}"
        self.log.append(msg)
        self.diffs_counter["missing_array_item"] += 1

    def _setup_summary(self) -> None:
        summary = (
            f"---------------------"
            f"\nTOTAL: {len(self.log)} differences\n"
        )
        for c_name, c_val in self.diffs_counter.items():
            if c_val:
                summary += f"-{c_name}: " + f"{c_val}\n"
        self.summary: str = summary
        self.log.append(summary)

    @staticmethod
    def __convert_to_json_type(
        item: int | float | str | bool | dict[str, Any] | list[Any] | None,
    ) -> str:
        if isinstance(item, int):
            return "<int>"
        elif isinstance(item, float):
            return "<float>"
        elif isinstance(item, str):
            return "<str>"
        elif isinstance(item, bool):
            return "<bool>"
        elif isinstance(item, dict):
            return "<object>"
        elif isinstance(item, list):
            return "<array>"
        else:
            return "<null>"
